#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/acct.h>
#include <ctype.h>
#include <signal.h>
#include "config.h"
#include "parser.h"
#include "inbuf.h"

extern char *cgetnametty(dev_t);
extern char *cgetnamuid(uid_t);
extern char *cgetnamgid(gid_t);
extern char *getelapsed(unsigned long);
extern char *getelapsed2(unsigned long);
extern char *gettimestr(struct timeval);

static char *sigtable[127];
static int sigtblinit = 0;

void
initsigtbl(void)
{
     sigtblinit = 1;

     memset(sigtable, 0, sizeof(sigtable));
     sigtable[SIGHUP] = "HUP";
     sigtable[SIGINT] = "INT";
     sigtable[SIGQUIT] = "QUIT";
     sigtable[SIGILL] = "ILL";
     sigtable[SIGABRT] = "ABRT";
     sigtable[SIGFPE] = "FPE";
     sigtable[SIGKILL] = "KILL";
     sigtable[SIGBUS] = "BUS";
     sigtable[SIGSEGV] = "SEGV";
     sigtable[SIGPIPE] = "PIPE";
     sigtable[SIGALRM] = "ALRM";
     sigtable[SIGTERM] = "TERM";
#ifdef SIGCHLD
     sigtable[SIGCHLD] = "CHLD";
#endif
#ifdef SIGCLD
     sigtable[SIGCLD] = "CHLD";
#endif
     sigtable[SIGSTOP] = "STOP";
#ifdef SIGXCPU
     sigtable[SIGXCPU] = "XCPU";
#endif
#ifdef SIGXFSZ
     sigtable[SIGXFSZ] = "XFSZ";
#endif
}

char *
fmtuser(struct acct *rec, int ignore)
{
     return cgetnamuid(rec->ac_uid);
}

char *
fmtgroup(struct acct *rec, int ignore)
{
     return cgetnamgid(rec->ac_gid);
}

char *
fmttty(struct acct *rec, int ignore)
{
     char *tty;

     tty = cgetnametty(rec->ac_tty);
     if(strncmp(tty, "/dev/", 5) == 0)
	  tty += 5;
     return tty;
}

char *
fmtstime(struct acct *rec, int which)
{
     switch(which){
     case 0:
	  return getelapsed2(rec->ac_utime);
     case 1:
	  return getelapsed2(rec->ac_stime);
     case 2:
	  return getelapsed(rec->ac_etime);
     default:
	  return "????";
     }
}

static struct tm *tm = 0;

char *
fmttime(struct acct *rec, int flag)
{
     struct timeval tmv;
     static char buf[80];
     char fmt[4];
     
     if(flag == 't'){
	  tmv.tv_sec = rec->ac_btime;
	  tmv.tv_usec = 0;
 
	  return gettimestr(tmv);
     }
     fmt[0] = '%';
     fmt[1] = (char)flag;
     fmt[2] = '\0';
     fmt[3] = '\0';

     if(!tm)
	  tm = localtime(&rec->ac_btime);

     strftime(buf, 80, fmt, tm);
     return buf;
}

char *
fmtcmd(struct acct *rec, int flag)
{
     static char cbuf[AC_COMMLEN+1];

     memcpy((void *)cbuf, (void *)rec->ac_comm, AC_COMMLEN);
     cbuf[AC_COMMLEN] = 0;
     return cbuf;
}

char *
fmtstatus(struct acct *rec, int flag)
{
     static char buf[4];

     buf[0] = rec->ac_flag & AFORK ? 'F' : ' ';
     buf[1] = rec->ac_flag & ASU ? 'P' : ' ';
     buf[2] = 0;

     return buf;
}

char *
fmtexitsig(struct acct *rec, int flag)
{
     static char buf[80];
     int signum;

     if(!sigtblinit)
	  initsigtbl();

     buf[0] = 0;
     if((signum = ACSIGEXIT(rec->ac_stat))){
	  if(sigtable[signum]){
	       return sigtable[signum];
	  }
	  else {
	       sprintf(buf, " SIG#%02d", signum);
	  }
     }
     return buf;
}
	  
#include "format.h"

void
doformat(struct acct *rec, struct args *ap, struct inbuf *buf)
{
     char *fmt;
     char fmtbuf[80];
     unsigned int i, ndx;
     int atflag;
     struct args *rove = ap->next;
     static int level = 1;

     tm = (struct tm *)0;

     for(fmt = ap->argp;*fmt;fmt++){
	  atflag = 0;
	  switch(*fmt){
	  case '@':
	       atflag=1;
	  case '%':
	       fmt++;
	       fmtbuf[0] = 0;
	       ndx = 0;
	       if(*fmt == '-'){
		    fmtbuf[ndx++] = '%';
		    fmtbuf[ndx++] = '-';
		    fmt++;
	       }
	       while(isdigit(*fmt) || *fmt == '.'){
		    if(!ndx)
			 fmtbuf[ndx++] = '%';
		    fmtbuf[ndx++] = *fmt++;
	       }
	       if(ndx){
		    fmtbuf[ndx++] = 's';
		    fmtbuf[ndx++] = '\0';
	       }
	       if(!atflag){
		    if(*fmt != 's'){
			 for(i=0;fmttbl[i].fmtstr;i++)
			      if(fmttbl[i].fmtchar == *fmt)
				   break;
			 if(fmttbl[i].fmtstr){
			      char *fmtstr = fmttbl[i].fmtstr;
			      if(fmtbuf[0])
				   fmtstr = fmtbuf;
			      addstrf(buf, fmtstr,
				     (*fmttbl[i].doit)(rec, fmttbl[i].flag));
			 }
			 else {
			      addchar(buf, '%');
			      addchar(buf, *fmt);
			 }
		    }
		    else { /* %s */
			 if(level > 1){
			      addstr(buf, "%s");
			 }
			 else {
			      static struct inbuf outbuf = { 0, 0, 0 };
			      char *fmtstr = "%s";

			      if(fmtbuf[0])
				   fmtstr = fmtbuf;
			      level++;
			      doformat(rec, rove, &outbuf);
			      addchar(&outbuf, '\0');
			      level--;
			      addstrf(buf, fmtstr, outbuf.buf);
			      resetbuf(&outbuf);
			 }
		    }
	       }
	       else {
		    if(fmtbuf[0])
			 addstrf(buf, fmtbuf, fmttime(rec, *fmt));
		    else
			 addstr(buf, fmttime(rec, *fmt));
	       }
	       break;
	  case '\\':
	       fmt++;
	       switch(*fmt){
	       case 'n':
		    addchar(buf, '\n');
		    break;
	       default:
		    addchar(buf, *fmt);
		    break;
	       }
	       break;
	  default:
	       addchar(buf, *fmt);
	       break;
	  }
     }
}

static struct inbuf output = { 0, 0, 0 };
static int linecount = 0;

void
flushoutput(FILE *logfile)
{
     write(fileno(logfile), output.buf, output.bufndx);
     resetbuf(&output);
     linecount = 0;
}

void
writeformatted(FILE *logfile, struct acct *rec, struct args *ap)
{
          
     doformat(rec, ap, &output);

     if(linecount++ > 500)
	  flushoutput(logfile);
}
     
